﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Data;
using System.Net;

namespace Qcew
{


    /// <summary>
    /// <para>Provides access to data from the Quarterly Census of Employment and Wages (QCEW).</para>
    /// <para>Provides data access by area, by industry, and data by establishment-size class.</para>
    /// <para>Data are returned in DataTables. (System.Data.DataTable) </para>
    /// </summary>
    public static class Data
    {



        //-----------------------------------------------------------------------------------------------------------------------------------
        #region GetAreaDataTable(year, qtr, area)

            /// <summary>
            /// <para>Gets a DataTable containing all records for a single area for a given time period. </para>
            /// </summary>
            /// <param name="year">String representing the year for the target time period</param>
            /// <param name="qtr">String representing the quarter for the target time period. Use "a" for annual averages. </param>
            /// <param name="area"><para>String representing the area fips of the target area. </para>
            /// <para>For a complete list of FIPS codes and titles, see http://data.bls.gov/cew/doc/titles/area/area_titles.htm </para>
            /// </param>
            /// <returns>DataTable containing all records for a single area for a given time period.</returns>
            public static DataTable GetAreaDataTable(string year, string qtr, string area)
            {
                var url = "http://data.bls.gov/cew/data/api/[YEAR]/[QTR]/area/[AREA].csv";
                url = url.Replace("[YEAR]", year);
                url = url.Replace("[QTR]", qtr.ToLower());
                url = url.Replace("[AREA]", area.ToUpper());
                var rows = GetAsRows(url);
                DataTable results;
                if (qtr == "a")
                {
                    results = ConvertRowsToDataTable(rows, GetEmptyAnnualTable());
                }
                else
                {
                    results = ConvertRowsToDataTable(rows, GetEmptyQuarterlyTable());
                }
                return results;
            }

        #endregion
        //-----------------------------------------------------------------------------------------------------------------------------------




        //-----------------------------------------------------------------------------------------------------------------------------------
        #region GetIndustryDataTable(year,qtr,industry)

            /// <summary>
            /// <para>Gets a DataTable containing all records for a specific industry for a given time period </para>
            /// </summary>
            /// <param name="year">String representing the year for the target time period </param>
            /// <param name="qtr">String representing the quarter for the target time period. Use "a" for annual averages. </param>
            /// <param name="industry"><para>String representing the industry code of the target industry.  </para>
            /// <para>NOTE: replace hyphens with underscores. "31-33" --> "31_33"</para>
            /// <para>For a complete list of industry codes and titles, see http://data.bls.gov/doc/titles/industry/industry_titles.htm </para>
            /// </param>
            /// <returns>DataTable containing all records for a single industry for the a given time period.</returns>
            public static DataTable GetIndustryDataTable(string year, string qtr, string industry)
            {
                var url = "http://data.bls.gov/cew/data/api/[YEAR]/[QTR]/industry/[INDUSTRY].csv";
                url = url.Replace("[YEAR]", year);
                url = url.Replace("[QTR]", qtr.ToLower());
                url = url.Replace("[INDUSTRY]", industry);
                var rows = GetAsRows(url);
                DataTable results;
                if (qtr == "a")
                {
                    results = ConvertRowsToDataTable(rows, GetEmptyAnnualTable());
                }
                else
                {
                    results = ConvertRowsToDataTable(rows, GetEmptyQuarterlyTable());
                }
                return results;
            }

        #endregion
        //-----------------------------------------------------------------------------------------------------------------------------------




        //-----------------------------------------------------------------------------------------------------------------------------------
        #region GetSizeDataTable(year,size)

            /// <summary>
            /// <para>Gets a DataTable containing all records for a specific establishment-size class, </para>
            /// <para>for the first quarter of a given year. Size data is only available for the first </para>
            /// <para>quarter of each year. </para>
            /// </summary>
            /// <param name="year">String representing the target year.</param>
            /// <param name="size"><para>String representing the target establishment-size class code.</para>
            /// <para>For a complete list of size-class codes and titles, see http://data.bls.gov/doc/titles/size/size_titles.htm </para>
            /// </param>
            /// <returns>DataTable containing all records for a single establishment-size class for a given year. </returns>
            public static DataTable GetSizeDataTable(string year, string size)
            {
                var url = "http://data.bls.gov/cew/data/api/[YEAR]/1/size/[SIZE].csv";
                url = url.Replace("[YEAR]", year);
                url = url.Replace("[SIZE]", size);
                var rows = GetAsRows(url);
                return ConvertRowsToDataTable(rows, GetEmptyQuarterlyTable());  //NOTE: Size data only available in first quarter.
            }

        #endregion
        //-----------------------------------------------------------------------------------------------------------------------------------




        //-----------------------------------------------------------------------------------------------------------------------------------
        #region PRIVATE GetEmptyQuarterlyTable() 
        
            /// <summary>
            /// <para>Returns an empty DataTable with columns for quarterly data. </para>
            /// </summary>
            /// <returns> Returns an empty DataTable with columns for quarterly data. </returns>
            private static DataTable GetEmptyQuarterlyTable()
            {
                DataTable table = new DataTable();
                Type Text = System.Type.GetType("System.String");
                Type Numeric = System.Type.GetType("System.Int64");
                Type DecimalNumber = System.Type.GetType("System.Decimal");
                table.Columns.Add("area_fips", Text);
                table.Columns.Add("own_code", Text);
                table.Columns.Add("industry_code", Text);
                table.Columns.Add("agglvl_code", Text);
                table.Columns.Add("size_code", Text);
                table.Columns.Add("year", Text);
                table.Columns.Add("qtr", Text);
                
                table.Columns.Add("disclosure_code", Text);
                table.Columns.Add("qtrly_estabs", Numeric);
                table.Columns.Add("month1_emplvl", Numeric);
                table.Columns.Add("month2_emplvl", Numeric);
                table.Columns.Add("month3_emplvl", Numeric);
                table.Columns.Add("total_qtrly_wages", Numeric);
                table.Columns.Add("taxable_qtrly_wages", Numeric);
                table.Columns.Add("qtrly_contributions", Numeric);
                table.Columns.Add("avg_wkly_wage", Numeric);
                
                table.Columns.Add("lq_disclosure_code", Text);
                table.Columns.Add("lq_qtrly_estabs", DecimalNumber);
                table.Columns.Add("lq_month1_emplvl", DecimalNumber);
                table.Columns.Add("lq_month2_emplvl", DecimalNumber);
                table.Columns.Add("lq_month3_emplvl", DecimalNumber);
                table.Columns.Add("lq_total_qtrly_wages", DecimalNumber);
                table.Columns.Add("lq_taxable_qtrly_wages", DecimalNumber);
                table.Columns.Add("lq_qtrly_contributions", DecimalNumber);
                table.Columns.Add("lq_avg_wkly_wage", DecimalNumber);
                
                table.Columns.Add("oty_disclosure_code", Text);
                table.Columns.Add("oty_qtrly_estabs_chg", Numeric);
                table.Columns.Add("oty_qtrly_estabs_pct_chg", DecimalNumber);
                table.Columns.Add("oty_month1_emplvl_chg", Numeric);
                table.Columns.Add("oty_month1_emplvl_pct_chg", DecimalNumber);
                table.Columns.Add("oty_month2_emplvl_chg", Numeric);
                table.Columns.Add("oty_month2_emplvl_pct_chg", DecimalNumber);
                table.Columns.Add("oty_month3_emplvl_chg", Numeric);
                table.Columns.Add("oty_month3_emplvl_pct_chg", DecimalNumber);
                table.Columns.Add("oty_total_qtrly_wages_chg", Numeric);
                table.Columns.Add("oty_total_qtrly_wages_pct_chg", DecimalNumber);
                table.Columns.Add("oty_taxable_qtrly_wages_chg", Numeric);
                table.Columns.Add("oty_taxable_qtrly_wages_pct_chg", DecimalNumber);
                table.Columns.Add("oty_qtrly_contributions_chg", Numeric);
                table.Columns.Add("oty_qtrly_contributions_pct_chg", DecimalNumber);
                table.Columns.Add("oty_avg_wkly_wage_chg", Numeric);
                table.Columns.Add("oty_avg_wkly_wage_pct_chg", DecimalNumber);

                
                return table;
            }
        
        #endregion
        //-----------------------------------------------------------------------------------------------------------------------------------




        //-----------------------------------------------------------------------------------------------------------------------------------
        #region PRIVATE GetEmptyAnnualTable()

            /// <summary>
            /// <para>Returns an empty DataTable with columns for annual data. </para>
            /// </summary>
            /// <returns>Returns an empty DataTable with columns for annual data </returns>
            private static DataTable GetEmptyAnnualTable()
            {
                DataTable table = new DataTable();
                Type Text = System.Type.GetType("System.String");
                Type Numeric = System.Type.GetType("System.Int64");
                Type DecimalNumber = System.Type.GetType("System.Decimal");
                table.Columns.Add("area_fips", Text);
                table.Columns.Add("own_code", Text);
                table.Columns.Add("industry_code", Text);
                table.Columns.Add("agglvl_code", Text);
                table.Columns.Add("size_code", Text);
                table.Columns.Add("year", Text);
                table.Columns.Add("qtr", Text);
                
                table.Columns.Add("disclosure_code", Text);
                table.Columns.Add("annual_avg_estabs", Numeric);
                table.Columns.Add("annual_avg_emplvl", Numeric);
                table.Columns.Add("total_annual_wages", Numeric);
                table.Columns.Add("taxable_annual_wages", Numeric);
                table.Columns.Add("annual_contributions", Numeric);
                table.Columns.Add("annual_avg_wkly_wage", Numeric);
                table.Columns.Add("avg_annual_pay", Numeric);
                
                table.Columns.Add("lq_disclosure_code", Text);
                table.Columns.Add("lq_annual_avg_estabs", DecimalNumber);
                table.Columns.Add("lq_annual_avg_emplvl", DecimalNumber);
                table.Columns.Add("lq_total_annual_wages", DecimalNumber);
                table.Columns.Add("lq_taxable_annual_wages", DecimalNumber);
                table.Columns.Add("lq_annual_contributions", DecimalNumber);
                table.Columns.Add("lq_annual_avg_wkly_wage", DecimalNumber);
                table.Columns.Add("lq_avg_annual_pay", DecimalNumber);

                table.Columns.Add("oty_disclosure_code", Text);
                table.Columns.Add("oty_annual_avg_estabs_chg", Numeric);
                table.Columns.Add("oty_annual_avg_estabs_pct_chg", DecimalNumber);
                table.Columns.Add("oty_annual_avg_emplvl_chg", Numeric);
                table.Columns.Add("oty_annual_avg_emplvl_pct_chg", DecimalNumber);
                table.Columns.Add("oty_total_annual_wages_chg", Numeric);
                table.Columns.Add("oty_total_annual_wages_pct_chg", DecimalNumber);
                table.Columns.Add("oty_taxable_annual_wages_chg", Numeric);
                table.Columns.Add("oty_taxable_annual_wages_pct_chg", DecimalNumber);
                table.Columns.Add("oty_annual_contributions_chg", Numeric);
                table.Columns.Add("oty_annual_contributions_pct_chg", DecimalNumber);
                table.Columns.Add("oty_annual_avg_wkly_wage_chg", Numeric);
                table.Columns.Add("oty_annual_avg_wkly_wage_pct_chg", DecimalNumber);
                table.Columns.Add("oty_avg_annual_pay_chg", Numeric);
                table.Columns.Add("oty_avg_annual_pay_pct_chg", DecimalNumber);

                return table;
            }

        #endregion
        //-----------------------------------------------------------------------------------------------------------------------------------




        //-----------------------------------------------------------------------------------------------------------------------------------
        #region PRIVATE ConvertRowsToDataTable(rows[], dataTable)

            /// <summary>
            /// Converts an array of Strings representing a CSV file into a DataTable. 
            /// </summary>
            /// <param name="rows">String Array representing the lines of the target CSV file.</param>
            /// <param name="dataTable">An empty DataTable with columns corresponding to the target CSV file.</param>
            /// <returns>A DataTable containing the CSV data.</returns>
            private static DataTable ConvertRowsToDataTable(string[] rows, DataTable dataTable)
            {
                string[] cells;
                foreach (var row in rows)
                {
                    cells = row.Replace("\"","").Split(',');   
                    dataTable.Rows.Add(cells);
                }
                return dataTable;
            }
        
        #endregion
        //-----------------------------------------------------------------------------------------------------------------------------------




        //-----------------------------------------------------------------------------------------------------------------------------------
        #region PRIVATE GetAsRows(csvUrl)
        
            /// <summary>
            /// Returns a string array representing the lines of a CSV file accessable via the web.
            /// </summary>
            /// <param name="csvUrl">a URL pointing to a CSV file available online.</param>
            /// <returns>A String Array representing lines of a CSV file.</returns>
            private static string[] GetAsRows(string csvUrl)
            {
                string html = "";
                using (var client = new WebClient())
                {
                    html = client.DownloadString(csvUrl);
                    html = html.Replace("\r", "");
                    html = html.Substring(html.IndexOf("\n") + 1); //remove headers
                }
                return html.Split(Convert.ToChar("\n"));
            }

        #endregion
        //-----------------------------------------------------------------------------------------------------------------------------------



    }
}
